@@ -16,9 +16,12 @@
 from ZODB.POSException import UndoError
 
 import logging
+import time
 
 log = logging.getLogger("relstorage.adapters.common")
 
+verify_sane_database = False
+
 
 # Notes about adapters:
 #
@@ -41,6 +44,7 @@ class Adapter(object):
         'TRUE':         'TRUE',
         'FALSE':        'FALSE',
         'OCTET_LENGTH': 'OCTET_LENGTH',
+        'TRUNCATE':     'TRUNCATE',
         'oid':          '%(oid)s',
         'tid':          '%(tid)s',
         'pack_tid':     '%(pack_tid)s',
@@ -51,16 +55,6 @@ class Adapter(object):
     }
 
     _scripts = {
-        'select_keep_tid': """
-            SELECT tid
-            FROM object_state
-            WHERE zoid = pack_object.zoid
-                AND tid > 0
-                AND tid <= %(pack_tid)s
-            ORDER BY tid DESC
-            LIMIT 1
-            """,
-
         'choose_pack_transaction': """
             SELECT tid
             FROM transaction
@@ -87,8 +81,59 @@ class Adapter(object):
             """,
 
         'reset_temp_undo': "DROP TABLE temp_undo",
-    }
 
+        'transaction_has_data': """
+            SELECT tid
+            FROM object_state
+            WHERE tid = %(tid)s
+            LIMIT 1
+            """,
+
+        'prepack_follow_child_refs': """
+            UPDATE pack_object SET keep = %(TRUE)s
+            WHERE keep = %(FALSE)s
+                AND zoid IN (
+                    SELECT DISTINCT to_zoid
+                    FROM object_ref
+                        JOIN temp_pack_visit USING (zoid)
+                )
+            """,
+
+        'pack_current_object': """
+            DELETE FROM current_object
+            WHERE tid = %(tid)s
+                AND zoid in (
+                    SELECT pack_state.zoid
+                    FROM pack_state
+                    WHERE pack_state.tid = %(tid)s
+                )
+            """,
+
+        'pack_object_state': """
+            DELETE FROM object_state
+            WHERE tid = %(tid)s
+                AND zoid in (
+                    SELECT pack_state.zoid
+                    FROM pack_state
+                    WHERE pack_state.tid = %(tid)s
+                )
+            """,
+
+        'pack_object_ref': """
+            DELETE FROM object_refs_added
+            WHERE tid IN (
+                SELECT tid
+                FROM transaction
+                WHERE empty = %(TRUE)s
+                );
+            DELETE FROM object_ref
+            WHERE tid IN (
+                SELECT tid
+                FROM transaction
+                WHERE empty = %(TRUE)s
+                )
+            """,
+    }
 
     def _run_script_stmt(self, cursor, generic_stmt, generic_params=()):
         """Execute a statement from a script with the given parameters.
@@ -137,7 +182,7 @@ class Adapter(object):
         and may have other columns.
         """
         for row in cursor:
-            tid, username, description = row[:3]
+            tid, username, description, ext = row[:4]
             if username is None:
                 username = ''
             else:
@@ -146,7 +191,11 @@ class Adapter(object):
                 description = ''
             else:
                 description = str(description)
-            yield (tid, username, description) + tuple(row[3:])
+            if ext is None:
+                ext = ''
+            else:
+                ext = str(ext)
+            yield (tid, username, description, ext) + tuple(row[4:])
 
 
     def iter_transactions(self, cursor):
@@ -170,7 +219,7 @@ class Adapter(object):
         """Iterate over the transactions in the given range, oldest first.
 
         Includes packed transactions.
-        Yields (tid, packed, username, description, extension)
+        Yields (tid, username, description, extension, packed)
         for each transaction.
         """
         stmt = """
@@ -358,77 +407,144 @@ class Adapter(object):
             self.close(conn, cursor)
 
 
-    def pre_pack(self, pack_tid, get_references, gc):
-        """Decide what to pack.
+    def open_for_pre_pack(self):
+        """Open a connection to be used for the pre-pack phase.
+        Returns (conn, cursor).
 
         Subclasses may override this.
+        """
+        return self.open()
+
+
+    def pre_pack(self, pack_tid, get_references, options):
+        """Decide what to pack.
 
         tid specifies the most recent transaction to pack.
 
         get_references is a function that accepts a pickled state and
         returns a set of OIDs that state refers to.
 
-        gc is a boolean indicating whether to run garbage collection.
-        If gc is false, at least one revision of every object is kept,
-        even if nothing refers to it.  Packing with gc disabled can be
+        options is an instance of relstorage.Options.
+        The options.pack_gc flag indicates whether to run garbage collection.
+        If pack_gc is false, at least one revision of every object is kept,
+        even if nothing refers to it.  Packing with pack_gc disabled can be
         much faster.
         """
-        conn, cursor = self.open()
+        conn, cursor = self.open_for_pre_pack()
         try:
             try:
-                if gc:
-                    self._pre_pack_with_gc(cursor, pack_tid, get_references)
+                if options.pack_gc:
+                    log.info("pre_pack: start with gc enabled")
+                    self._pre_pack_with_gc(
+                        conn, cursor, pack_tid, get_references)
                 else:
-                    self._pre_pack_without_gc(cursor, pack_tid)
+                    log.info("pre_pack: start without gc")
+                    self._pre_pack_without_gc(
+                        conn, cursor, pack_tid)
+                conn.commit()
+
+                log.info("pre_pack: enumerating states to pack")
+                stmt = "%(TRUNCATE)s pack_state"
+                self._run_script_stmt(cursor, stmt)
+                to_remove = 0
+
+                if options.pack_gc:
+                    # Pack objects with the keep flag set to false.
+                    stmt = """
+                    INSERT INTO pack_state (tid, zoid)
+                    SELECT tid, zoid
+                    FROM object_state
+                        JOIN pack_object USING (zoid)
+                    WHERE keep = %(FALSE)s
+                        AND tid > 0
+                        AND tid <= %(pack_tid)s
+                    """
+                    self._run_script_stmt(cursor, stmt, {'pack_tid': pack_tid})
+                    to_remove += cursor.rowcount
+
+                # Pack object states with the keep flag set to true.
+                stmt = """
+                INSERT INTO pack_state (tid, zoid)
+                SELECT tid, zoid
+                FROM object_state
+                    JOIN pack_object USING (zoid)
+                WHERE keep = %(TRUE)s
+                    AND tid > 0
+                    AND tid != keep_tid
+                    AND tid <= %(pack_tid)s
+                """
+                self._run_script_stmt(cursor, stmt, {'pack_tid':pack_tid})
+                to_remove += cursor.rowcount
+
+                log.info("pre_pack: enumerating transactions to pack")
+                stmt = "%(TRUNCATE)s pack_state_tid"
+                self._run_script_stmt(cursor, stmt)
+                stmt = """
+                INSERT INTO pack_state_tid (tid)
+                SELECT DISTINCT tid
+                FROM pack_state
+                """
+                cursor.execute(stmt)
+
+                log.info("pre_pack: will remove %d object state(s)",
+                    to_remove)
+
             except:
+                log.exception("pre_pack: failed")
                 conn.rollback()
                 raise
             else:
+                log.info("pre_pack: finished successfully")
                 conn.commit()
         finally:
             self.close(conn, cursor)
 
 
-    def _pre_pack_without_gc(self, cursor, pack_tid):
+    def _pre_pack_without_gc(self, conn, cursor, pack_tid):
         """Determine what to pack, without garbage collection.
 
         With garbage collection disabled, there is no need to follow
         object references.
         """
         # Fill the pack_object table with OIDs, but configure them
-        # all to be kept by setting keep and keep_tid.
+        # all to be kept by setting keep to true.
+        log.debug("pre_pack: populating pack_object")
         stmt = """
-        DELETE FROM pack_object;
+        %(TRUNCATE)s pack_object;
 
-        INSERT INTO pack_object (zoid, keep)
-        SELECT DISTINCT zoid, %(TRUE)s
+        INSERT INTO pack_object (zoid, keep, keep_tid)
+        SELECT zoid, %(TRUE)s, MAX(tid)
         FROM object_state
-        WHERE tid <= %(pack_tid)s;
-
-        UPDATE pack_object SET keep_tid = (@select_keep_tid@)
+        WHERE tid > 0 AND tid <= %(pack_tid)s
+        GROUP BY zoid
         """
-        stmt = stmt.replace(
-            '@select_keep_tid@', self._scripts['select_keep_tid'])
         self._run_script(cursor, stmt, {'pack_tid': pack_tid})
 
 
-    def _pre_pack_with_gc(self, cursor, pack_tid, get_references):
+    def _pre_pack_with_gc(self, conn, cursor, pack_tid, get_references):
         """Determine what to pack, with garbage collection.
         """
+        stmt = self._scripts['create_temp_pack_visit']
+        if stmt:
+            self._run_script(cursor, stmt)
+
+        log.info("pre_pack: following references after the pack point")
         # Fill object_ref with references from object states
         # in transactions that will not be packed.
-        self._fill_nonpacked_refs(cursor, pack_tid, get_references)
+        self._fill_nonpacked_refs(conn, cursor, pack_tid, get_references)
 
+        log.debug("pre_pack: populating pack_object")
         # Fill the pack_object table with OIDs that either will be
         # removed (if nothing references the OID) or whose history will
         # be cut.
         stmt = """
-        DELETE FROM pack_object;
+        %(TRUNCATE)s pack_object;
 
-        INSERT INTO pack_object (zoid, keep)
-        SELECT DISTINCT zoid, %(FALSE)s
+        INSERT INTO pack_object (zoid, keep, keep_tid)
+        SELECT zoid, %(FALSE)s, MAX(tid)
         FROM object_state
-        WHERE tid <= %(pack_tid)s;
+        WHERE tid > 0 AND tid <= %(pack_tid)s
+        GROUP BY zoid;
 
         -- If the root object is in pack_object, keep it.
         UPDATE pack_object SET keep = %(TRUE)s
@@ -436,77 +552,93 @@ class Adapter(object):
 
         -- Keep objects that have been revised since pack_tid.
         UPDATE pack_object SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND zoid IN (
-                SELECT zoid
-                FROM current_object
-                WHERE tid > %(pack_tid)s
-            );
+        WHERE zoid IN (
+            SELECT zoid
+            FROM current_object
+            WHERE tid > %(pack_tid)s
+        );
 
         -- Keep objects that are still referenced by object states in
         -- transactions that will not be packed.
+        -- Use temp_pack_visit for temporary state; otherwise MySQL 5 chokes.
+        INSERT INTO temp_pack_visit
+        SELECT DISTINCT to_zoid
+        FROM object_ref
+        WHERE tid > %(pack_tid)s;
+
         UPDATE pack_object SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND zoid IN (
-                SELECT to_zoid
-                FROM object_ref
-                WHERE tid > %(pack_tid)s
-            );
+        WHERE zoid IN (
+            SELECT zoid
+            FROM temp_pack_visit
+        );
+
+        %(TRUNCATE)s temp_pack_visit;
         """
         self._run_script(cursor, stmt, {'pack_tid': pack_tid})
 
-        stmt = self._scripts['create_temp_pack_visit']
-        if stmt:
-            self._run_script(cursor, stmt)
-
         # Each of the packable objects to be kept might
         # refer to other objects.  If some of those references
-        # include objects currently set to be removed, keep
-        # those objects as well.  Do this
+        # include objects currently set to be removed, mark
+        # the referenced objects to be kept as well.  Do this
         # repeatedly until all references have been satisfied.
+        pass_num = 1
         while True:
+            log.info("pre_pack: following references before the pack point, "
+                "pass %d", pass_num)
 
             # Make a list of all parent objects that still need
-            # to be visited.  Then set keep_tid for all pack_object
+            # to be visited.  Then set pack_object.visited for all pack_object
             # rows with keep = true.
-            # keep_tid must be set before _fill_pack_object_refs examines
-            # references.
             stmt = """
-            DELETE FROM temp_pack_visit;
+            %(TRUNCATE)s temp_pack_visit;
 
             INSERT INTO temp_pack_visit (zoid)
             SELECT zoid
             FROM pack_object
             WHERE keep = %(TRUE)s
-                AND keep_tid IS NULL;
+                AND visited = %(FALSE)s;
 
-            UPDATE pack_object SET keep_tid = (@select_keep_tid@)
-            WHERE keep = %(TRUE)s AND keep_tid IS NULL
+            UPDATE pack_object SET visited = %(TRUE)s
+            WHERE keep = %(TRUE)s
+                AND visited = %(FALSE)s
             """
-            stmt = stmt.replace(
-                '@select_keep_tid@', self._scripts['select_keep_tid'])
-            self._run_script(cursor, stmt, {'pack_tid': pack_tid})
+            self._run_script(cursor, stmt)
+            visit_count = cursor.rowcount
 
-            self._fill_pack_object_refs(cursor, get_references)
+            if verify_sane_database:
+                # Verify the update actually worked.
+                # MySQL 5.1.23 fails this test; 5.1.24 passes.
+                stmt = """
+                SELECT 1
+                FROM pack_object
+                WHERE keep = %(TRUE)s AND visited = %(FALSE)s
+                """
+                self._run_script_stmt(cursor, stmt)
+                if list(cursor):
+                    raise AssertionError(
+                        "database failed to update pack_object")
+
+            log.debug("pre_pack: checking references from %d object(s)",
+                visit_count)
+
+            self._fill_pack_object_refs(conn, cursor, get_references)
 
             # Visit the children of all parent objects that were
             # just visited.
-            stmt = """
-            UPDATE pack_object SET keep = %(TRUE)s
-            WHERE keep = %(FALSE)s
-                AND zoid IN (
-                    SELECT DISTINCT to_zoid
-                    FROM object_ref
-                        JOIN temp_pack_visit USING (zoid)
-                )
-            """
-            self._run_script_stmt(cursor, stmt)
-            if not cursor.rowcount:
+            stmt = self._scripts['prepack_follow_child_refs']
+            self._run_script(cursor, stmt)
+            found_count = cursor.rowcount
+
+            log.debug("pre_pack: found %d more referenced object(s) in "
+                "pass %d", found_count, pass_num)
+            if not found_count:
                 # No new references detected.
                 break
+            else:
+                pass_num += 1
 
 
-    def _fill_nonpacked_refs(self, cursor, pack_tid, get_references):
+    def _fill_nonpacked_refs(self, conn, cursor, pack_tid, get_references):
         """Fill object_ref for all transactions that will not be packed."""
         stmt = """
         SELECT DISTINCT tid
@@ -519,25 +651,22 @@ class Adapter(object):
             )
         """
         self._run_script_stmt(cursor, stmt, {'pack_tid': pack_tid})
-        for (tid,) in cursor.fetchall():
-            self._add_refs_for_tid(cursor, tid, get_references)
+        tids = [tid for (tid,) in cursor]
+        self._add_refs_for_tids(conn, cursor, tids, get_references)
 
 
-    def _fill_pack_object_refs(self, cursor, get_references):
-        """Fill object_ref for all pack_object rows that have keep_tid."""
+    def _fill_pack_object_refs(self, conn, cursor, get_references):
+        """Fill object_ref for all objects that are to be kept."""
         stmt = """
         SELECT DISTINCT keep_tid
         FROM pack_object
-        WHERE keep_tid IS NOT NULL
-            AND NOT EXISTS (
-                SELECT 1
-                FROM object_refs_added
-                WHERE tid = keep_tid
-            )
+            LEFT JOIN object_refs_added ON (keep_tid = tid)
+        WHERE keep = %(TRUE)s
+            AND object_refs_added.tid IS NULL
         """
-        cursor.execute(stmt)
-        for (tid,) in cursor.fetchall():
-            self._add_refs_for_tid(cursor, tid, get_references)
+        self._run_script_stmt(cursor, stmt)
+        tids = [tid for (tid,) in cursor]
+        self._add_refs_for_tids(conn, cursor, tids, get_references)
 
 
     def _add_object_ref_rows(self, cursor, add_rows):
@@ -556,7 +685,12 @@ class Adapter(object):
 
     def _add_refs_for_tid(self, cursor, tid, get_references):
         """Fill object_refs with all states for a transaction.
+
+        Returns the number of references added.
         """
+        log.debug("pre_pack: transaction %d: computing references ", tid)
+        from_count = 0
+
         stmt = """
         SELECT zoid, state
         FROM object_state
@@ -570,6 +704,7 @@ class Adapter(object):
                 # Oracle
                 state = state.read()
             if state:
+                from_count += 1
                 to_oids = get_references(str(state))
                 for to_oid in to_oids:
                     add_rows.append((from_oid, tid, to_oid))
@@ -584,103 +719,182 @@ class Adapter(object):
         """
         self._run_script_stmt(cursor, stmt, {'tid': tid})
 
+        to_count = len(add_rows)
+        log.debug("pre_pack: transaction %d: has %d reference(s) "
+            "from %d object(s)", tid, to_count, from_count)
+        return to_count
+
+
+    def _add_refs_for_tids(self, conn, cursor, tids, get_references):
+        """Fill object_refs with all states for multiple transactions."""
+        if tids:
+            added = 0
+            log.info("pre_pack: discovering references from objects in %d "
+                "transaction(s)" % len(tids))
+            for tid in tids:
+                added += self._add_refs_for_tid(cursor, tid, get_references)
+                if added >= 1000:
+                    # save the work done so far
+                    conn.commit()
+                    added = 0
+
 
     def _hold_commit_lock(self, cursor):
         """Hold the commit lock for packing"""
         cursor.execute("LOCK TABLE commit_lock IN EXCLUSIVE MODE")
 
 
-    def pack(self, pack_tid):
-        """Pack.  Requires populated pack tables."""
+    def _release_commit_lock(self, cursor):
+        """Release the commit lock during packing"""
+        # no action needed
+        pass
+
+
+    def pack(self, pack_tid, options):
+        """Pack.  Requires the information provided by pre_pack."""
 
         # Read committed mode is sufficient.
         conn, cursor = self.open()
         try:
             try:
-                # hold the commit lock for a moment to prevent deadlocks.
-                self._hold_commit_lock(cursor)
-
-                for table in ('object_ref', 'current_object', 'object_state'):
-
-                    # Remove objects that are in pack_object and have keep
-                    # set to false.
-                    stmt = """
-                    DELETE FROM %s
-                    WHERE zoid IN (
-                            SELECT zoid
-                            FROM pack_object
-                            WHERE keep = %%(FALSE)s
-                        )
-                    """ % table
-                    self._run_script_stmt(cursor, stmt)
-
-                    if table != 'current_object':
-                        # Cut the history of objects in pack_object that
-                        # have keep set to true.
-                        stmt = """
-                        DELETE FROM %s
-                        WHERE zoid IN (
-                                SELECT zoid
-                                FROM pack_object
-                                WHERE keep = %%(TRUE)s
-                            )
-                            AND tid < (
-                                SELECT keep_tid
-                                FROM pack_object
-                                WHERE zoid = %s.zoid
-                            )
-                        """ % (table, table)
-                        self._run_script_stmt(cursor, stmt)
-
                 stmt = """
-                -- Terminate prev_tid chains
-                UPDATE object_state SET prev_tid = 0
-                WHERE tid <= %(pack_tid)s
-                    AND prev_tid != 0;
-
-                -- For each tid to be removed, delete the corresponding row in
-                -- object_refs_added.
-                DELETE FROM object_refs_added
-                WHERE tid > 0
-                    AND tid <= %(pack_tid)s
-                    AND NOT EXISTS (
-                        SELECT 1
-                        FROM object_state
-                        WHERE tid = object_refs_added.tid
-                    );
-
-                -- Delete transactions no longer used.
-                DELETE FROM transaction
-                WHERE tid > 0
-                    AND tid <= %(pack_tid)s
-                    AND NOT EXISTS (
-                        SELECT 1
-                        FROM object_state
-                        WHERE tid = transaction.tid
-                    );
-
-                -- Mark the remaining packable transactions as packed
-                UPDATE transaction SET packed = %(TRUE)s
-                WHERE tid > 0
-                    AND tid <= %(pack_tid)s
-                    AND packed = %(FALSE)s;
-
-                -- Clean up.
-                DELETE FROM pack_object;
+                SELECT transaction.tid,
+                    CASE WHEN packed = %(TRUE)s THEN 1 ELSE 0 END,
+                    CASE WHEN pack_state_tid.tid IS NOT NULL THEN 1 ELSE 0 END
+                FROM transaction
+                    LEFT JOIN pack_state_tid ON (
+                        transaction.tid = pack_state_tid.tid)
+                WHERE transaction.tid > 0
+                    AND transaction.tid <= %(pack_tid)s
+                    AND (packed = %(FALSE)s OR pack_state_tid.tid IS NOT NULL)
                 """
-                self._run_script(cursor, stmt, {'pack_tid': pack_tid})
+                self._run_script_stmt(cursor, stmt, {'pack_tid': pack_tid})
+                tid_rows = list(cursor)
+                tid_rows.sort()  # oldest first
+
+                log.info("pack: will pack %d transaction(s)", len(tid_rows))
+
+                stmt = self._scripts['create_temp_pack_visit']
+                if stmt:
+                    self._run_script(cursor, stmt)
+
+                # Hold the commit lock while packing to prevent deadlocks.
+                # Pack in small batches of transactions in order to minimize
+                # the interruption of concurrent write operations.
+                start = time.time()
+                self._hold_commit_lock(cursor)
+                for tid, packed, has_removable in tid_rows:
+                    self._pack_transaction(
+                        cursor, pack_tid, tid, packed, has_removable)
+                    if time.time() >= start + options.pack_batch_timeout:
+                        # commit the work done so far and release the
+                        # commit lock for a short time
+                        conn.commit()
+                        self._release_commit_lock(cursor)
+                        # Add a delay based on the configured duty cycle.
+                        elapsed = time.time() - start
+                        duty_cycle = options.pack_duty_cycle
+                        if duty_cycle > 0.0 and duty_cycle < 1.0:
+                            delay = min(options.pack_max_delay,
+                                elapsed * (1.0 / duty_cycle - 1.0))
+                            if delay > 0:
+                                log.debug('pack: sleeping %.4g second(s)',
+                                    delay)
+                                time.sleep(delay)
+                        self._hold_commit_lock(cursor)
+                        start = time.time()
+
+                self._pack_cleanup(conn, cursor)
 
             except:
+                log.exception("pack: failed")
                 conn.rollback()
                 raise
 
             else:
+                log.info("pack: finished successfully")
                 conn.commit()
 
         finally:
             self.close(conn, cursor)
 
 
+    def _pack_transaction(self, cursor, pack_tid, tid, packed,
+            has_removable):
+        """Pack one transaction.  Requires populated pack tables."""
+        log.debug("pack: transaction %d: packing", tid)
+        removed_objects = 0
+        removed_states = 0
+
+        if has_removable:
+            stmt = self._scripts['pack_current_object']
+            self._run_script_stmt(cursor, stmt, {'tid': tid})
+            removed_objects = cursor.rowcount
+
+            stmt = self._scripts['pack_object_state']
+            self._run_script_stmt(cursor, stmt, {'tid': tid})
+            removed_states = cursor.rowcount
+
+            # Terminate prev_tid chains
+            stmt = """
+            UPDATE object_state SET prev_tid = 0
+            WHERE prev_tid = %(tid)s
+                AND tid <= %(pack_tid)s
+            """
+            self._run_script_stmt(cursor, stmt,
+                {'pack_tid': pack_tid, 'tid': tid})
+
+        # Find out whether the transaction is empty
+        stmt = self._scripts['transaction_has_data']
+        self._run_script_stmt(cursor, stmt, {'tid': tid})
+        empty = not list(cursor)
+
+        # mark the transaction packed and possibly empty
+        if empty:
+            clause = 'empty = %(TRUE)s'
+            state = 'empty'
+        else:
+            clause = 'empty = %(FALSE)s'
+            state = 'not empty'
+        stmt = "UPDATE transaction SET packed = %(TRUE)s, " + clause
+        stmt += " WHERE tid = %(tid)s"
+        self._run_script_stmt(cursor, stmt, {'tid': tid})
+
+        log.debug(
+            "pack: transaction %d (%s): removed %d object(s) and %d state(s)",
+            tid, state, removed_objects, removed_states)
+
+
+    def _pack_cleanup(self, conn, cursor):
+        """Remove unneeded table rows after packing"""
+        # commit the work done so far
+        conn.commit()
+        self._release_commit_lock(cursor)
+        self._hold_commit_lock(cursor)
+        log.info("pack: cleaning up")
+
+        log.debug("pack: removing unused object references")
+        stmt = self._scripts['pack_object_ref']
+        self._run_script(cursor, stmt)
+
+        log.debug("pack: removing empty packed transactions")
+        stmt = """
+        DELETE FROM transaction
+        WHERE packed = %(TRUE)s
+            AND empty = %(TRUE)s
+        """
+        self._run_script_stmt(cursor, stmt)
+
+        # perform cleanup that does not require the commit lock
+        conn.commit()
+        self._release_commit_lock(cursor)
+
+        log.debug("pack: clearing temporary pack state")
+        for _table in ('pack_object', 'pack_state', 'pack_state_tid'):
+            stmt = '%(TRUNCATE)s ' + _table
+            self._run_script_stmt(cursor, stmt)
+
+
     def poll_invalidations(self, conn, cursor, prev_polled_tid, ignore_tid):
         """Polls for new transactions.
 
@@ -705,7 +919,8 @@ class Adapter(object):
             return (), new_polled_tid
 
         stmt = "SELECT 1 FROM transaction WHERE tid = %(tid)s"
-        cursor.execute(stmt % self._script_vars, {'tid': prev_polled_tid})
+        cursor.execute(intern(stmt % self._script_vars),
+            {'tid': prev_polled_tid})
         rows = cursor.fetchall()
         if not rows:
             # Transaction not found; perhaps it has been packed.
@@ -714,15 +929,16 @@ class Adapter(object):
 
         # Get the list of changed OIDs and return it.
         stmt = """
-        SELECT DISTINCT zoid
-        FROM object_state
+        SELECT zoid
+        FROM current_object
         WHERE tid > %(tid)s
         """
         if ignore_tid is None:
-            cursor.execute(stmt % self._script_vars, {'tid': prev_polled_tid})
+            cursor.execute(intern(stmt % self._script_vars),
+                {'tid': prev_polled_tid})
         else:
             stmt += " AND tid != %(self_tid)s"
-            cursor.execute(stmt % self._script_vars,
+            cursor.execute(intern(stmt % self._script_vars),
                 {'tid': prev_polled_tid, 'self_tid': ignore_tid})
         oids = [oid for (oid,) in cursor]
 
